@@ -9,7 +9,7 @@
 
 require 5.004;
 
-$DBD::ODBC::VERSION = '0.33_3';
+$DBD::ODBC::VERSION = '0.45_4';
 
 {
     package DBD::ODBC;
@@ -59,7 +59,7 @@ $DBD::ODBC::VERSION = '0.33_3';
 
     sub connect {
 	my $drh = shift;
-	my($dbname, $user, $auth)= @_;
+	my($dbname, $user, $auth, $attr)= @_;
 	$user = '' unless defined $user;
 	$auth = '' unless defined $auth;
 
@@ -73,7 +73,7 @@ $DBD::ODBC::VERSION = '0.33_3';
 	# Call ODBC logon func in ODBC.xs file
 	# and populate internal handle data.
 
-	DBD::ODBC::db::_login($this, $dbname, $user, $auth) or return undef;
+	DBD::ODBC::db::_login($this, $dbname, $user, $auth, $attr) or return undef;
 
 	$this;
     }
@@ -102,10 +102,29 @@ $DBD::ODBC::VERSION = '0.33_3';
 	$sth;
     }
 
+    sub column_info {
+	my ($dbh, $catalog, $schema, $table, $column) = @_;
 
+	$catalog = "" if (!$catalog);
+	$schema = "" if (!$schema);
+	$table = "" if (!$table);
+	$column = "" if (!$column);
+	# create a "blank" statement handle
+	my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLColumns" });
+
+	_columns($dbh,$sth, $catalog, $schema, $table, $column)
+	    or return undef;
+
+	$sth;
+    }
+    
     sub columns {
 	my ($dbh, $catalog, $schema, $table, $column) = @_;
 
+	$catalog = "" if (!$catalog);
+	$schema = "" if (!$schema);
+	$table = "" if (!$table);
+	$column = "" if (!$column);
 	# create a "blank" statement handle
 	my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLColumns" });
 
@@ -117,21 +136,85 @@ $DBD::ODBC::VERSION = '0.33_3';
 
 
     sub table_info {
-	my($dbh) = @_;		# XXX add qualification
+ 	my($dbh, $catalog, $schema, $table, $type) = @_;
+
+	if ($#_ == 1) {
+	   my $attrs = $_[1];
+	   $catalog = $attrs->{TABLE_CAT};
+	   $schema = $attrs->{TABLE_SCHEM};
+	   $table = $attrs->{TABLE_NAME};
+	   $type = $attrs->{TABLE_TYPE};
+ 	}
+
+	$catalog = "" if (!$catalog);
+	$schema = "" if (!$schema);
+	$table = "" if (!$table);
+	$type = "" if (!$type);
 
 	# create a "blank" statement handle
 	my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLTables" });
 
-	# XXX use qualification(s) (qual, schema, etc?) here...
-	DBD::ODBC::st::_tables($dbh,$sth, "")
-		or return undef;
+	DBD::ODBC::st::_tables($dbh,$sth, $catalog, $schema, $table, $type)
+	      or return undef;
 	$sth;
     }
 
+    sub primary_key_info {
+       my ($dbh, $catalog, $schema, $table ) = @_;
+ 
+       # create a "blank" statement handle
+       my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLPrimaryKeys" });
+ 
+       $catalog = "" if (!$catalog);
+       $schema = "" if (!$schema);
+       $table = "" if (!$table);
+       DBD::ODBC::st::_primary_keys($dbh,$sth, $catalog, $schema, $table )
+	     or return undef;
+       $sth;
+    }
+
+    sub foreign_key_info {
+       my ($dbh, $pkcatalog, $pkschema, $pktable, $fkcatalog, $fkschema, $fktable ) = @_;
+ 
+       # create a "blank" statement handle
+       my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLForeignKeys" });
+ 
+       $pkcatalog = "" if (!$pkcatalog);
+       $pkschema = "" if (!$pkschema);
+       $pktable = "" if (!$pktable);
+       $fkcatalog = "" if (!$fkcatalog);
+       $fkschema = "" if (!$fkschema);
+       $fktable = "" if (!$fktable);
+       _GetForeignKeys($dbh, $sth, $pkcatalog, $pkschema, $pktable, $fkcatalog, $fkschema, $fktable) or return undef;
+       $sth;
+    }
 
     sub ping {
 	my $dbh = shift;
 	my $state = undef;
+
+ 	my ($catalog, $schema, $table, $type);
+
+	$catalog = "";
+	$schema = "";
+	$table = "NOXXTABLE";
+	$type = "";
+
+	# create a "blank" statement handle
+	my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLTables_PING" });
+
+	DBD::ODBC::st::_tables($dbh,$sth, $catalog, $schema, $table, $type)
+	      or return 0;
+	$sth->finish;
+	return 1;
+
+    }
+
+    # saved, just for posterity.
+    sub oldping  {
+	my $dbh = shift;
+	my $state = undef;
+
 	# should never 'work' but if it does, that's okay!
 	# JLU incorporated patches from Jon Smirl 5/4/99
 	{
@@ -159,12 +242,62 @@ $DBD::ODBC::VERSION = '0.33_3';
  	return 1 if $state eq '42S02';  # Base table not found.Solid EE v3.51
 	return 1 if $state eq 'S0022';	# Column not found
 	return 1 if $state eq '37000';  # statement could not be prepared (19991011, JLU)
+	# return 1 if $state eq 'S1000';  # General Error? ? 5/30/02, JLU.  This is what Openlink is returning
 	# We assume that any other error means the database
 	# is no longer connected.
 	# Some special cases may need to be added to the code above.
 	return 0;
     }
 
+    # New support for the next DBI which will have a get_info command.
+    # leaving support for ->func(xxx, GetInfo) (above) for a period of time
+    # to support older applications which used this.
+    sub get_info {
+	my ($dbh, $item) = @_;
+	# handle SQL_DRIVER_HSTMT, SQL_DRIVER_HLIB and
+	# SQL_DRIVER_HDESC specially
+	if ($item == 5 || $item == 135 || $item == 76) {
+	   return undef;
+	}
+	return _GetInfo($dbh, $item);
+    }
+
+    # new override of do method provided by Merijn Broeren
+    # this optimizes "do" to use SQLExecDirect for simple
+    # do statements without parameters.
+    sub do {
+        my($dbh, $statement, $attr, @params) = @_;
+        my $rows = 0;
+
+        if( -1 == $#params )
+        {
+          # No parameters, use execute immediate
+          $rows = ExecDirect( $dbh, $statement );
+          if( 0 == $rows )
+          {
+            $rows = "0E0";
+          }
+          elsif( $rows < -1 )
+          {
+            undef $rows;
+          }
+        }
+        else
+        {
+          $rows = $dbh->SUPER::do( $statement, $attr, @params );
+        }
+        return $rows
+    }
+
+    #
+    # can also be called as $dbh->func($sql, ExecDirect);
+    # if, for some reason, there are compatibility issues
+    # later with DBI's do.
+    #
+    sub ExecDirect {
+       my ($dbh, $sql) = @_;
+       _ExecDirect($dbh, $sql);
+    }
 
     # Call the ODBC function SQLGetInfo
     # Args are:
@@ -174,7 +307,7 @@ $DBD::ODBC::VERSION = '0.33_3';
     #
     sub GetInfo {
 	my ($dbh, $item) = @_;
-	_GetInfo($dbh, $item);
+	get_info($dbh, $item);
     }
 
     # Call the ODBC function SQLStatistics
@@ -287,7 +420,199 @@ See L<DBI> for more information.
 =head2 Recent Updates
 
 =over 4
+=item B<An Important note about the tests!>
 
+ Please note that some tests may fail or report they are
+ unsupported on this platform.  Notably Oracle's ODBC driver
+ will fail the "advanced" binding tests in t/08bind2.t.
+ These tests run perfectly under SQL Server 2000. This is
+ normal and expected.  Until Oracle fixes their drivers to
+ do the right thing from an ODBC perspective, it's going to
+ be tough to fix the issue.  The workaround for Oracle is to
+ bind date types with SQL_TIMESTAMP.
+   
+ Also note that some tests may be skipped, such as
+ t/09multi.t, if your driver doesn't seem to support
+ returning multiple result sets.
+
+=item B<DBD::ODBC 0.45_4>
+
+More fixes for DB2 tests and timestamp handling.
+ 
+=item B<DBD::ODBC 0.45_3>
+
+Changes to internal timestamp type handling and test structure to ensure tests
+work for all platforms.  DB2 was giving me fits due to bad assumptions.  Thanks
+to Martin Evans (again) for help in identifying the problems and helping research
+solutions.  This includes the scale/precision values to correctly store full timestamps.
+ 
+=item B<DBD::ODBC 0.45_2>
+
+Moving API usage to ODBC 3.0 specifications.  With lots of help from Martin Evans (again!).
+Thanks Martin!!!!!
+ 
+=item B<DBD::ODBC 0.44>
+
+Fix for do() and execute to handle DB2 correctly.  Patch/discovery thanks to Martin Evans.
+Partly moving towards defaulting to ODBC 3.x standards.
+
+=item B<DBD::ODBC 0.43>
+
+Fix for FoxPro drivers -- HURRAY!!!!!!!!!  Bad call to SQLDescribeCol where the FoxPro driver
+seems to clear out the buffer passed and the size of the buffer indicated in the parameters
+was larger than the real buffer...need better fix long term, but fixed for now by allocating
+extra space.
+ 
+Added support for column_info (maps to function columns, previously provided)
+ 
+Fix for binding undef value generated by a dereferencing an empty hash value.
+I.e in the hash, there is no key value "1", but $hash{"1"} is passed to binding parameters.
+
+A fix to make all bound columns word aligned.  This solved a problem on solaris when binding
+timestamp values after a varchar value.  This patch forces all columns to start on word
+boundaries (sizeof int boundaries, actually).  This may have a positive speed impact for
+other systems.  Thanks to Joe Slagel for pointing this out and providing a patch for the fix!
+ 
+=item B<DBD::ODBC 0.42>
+
+A patch to the tests to get past any potential problems with
+a user not defining the DBI_DSN environment variable.  Hopefully
+this will allow ActiveState to keep up with DBD::ODBC in
+their automated builds.  Thanks to Jan Dubois for helping
+me determine what the issue(s) may be when testing.
+Now skips all tests except the dll/so load if DBI_DSN is
+not defined.  Also, issues a skip on a known Oracle ODBC
+test.
+ 
+A patch to fix ping() when a S1000 was returned.  I don't know how this
+one is going to go, but it's worth a shot.  Now calls SQLTables with unlikely
+parameters.  If SQLTables() succeeds and returns a result set, ping() will
+now return true.  Thanks to Tim Bunce for the suggestion.
+ 
+=item B<DBD::ODBC 0.41>
+
+A patch to handle SQLDescribeParam failing in some
+circumstances.  I believe this is a bug in the ODBC
+driver, but this happens with SQL Server and INNER
+JOIN syntax queries, but not when where a.i = b.i.
+The behavior is now to revert to assume SQL_VARCHAR
+if SQLDescribeParam fails for whatever reason.					     
+
+Added warning during tests when the Oracle ODBC driver
+is detected and it is a known error.  I didn't want
+to make it pass the test silently, but I did want to print
+a message indicating it was expected.
+   
+=item B<DBD::ODBC 0.40>
+
+Two minor patches for building.  One for Cygwin support
+and another to handle the case where both iODBC and
+unixODBC libraries are installed.  The preference is
+now given to unixODBC...  
+
+Fixed problem in connect code introduced in 0.39 where
+new code would only be executed if tracing was on and
+> 3...
+
+NEW: Changed default bind type to 0, to let DBD::ODBC
+try to detect the correct bind type.  This is a
+potentially significant change.  Please let me know
+if there is a problem.  The best way is to post a message
+to the dbi-users@perl.org mailing list.
+
+=item B<DBD::ODBC 0.39>
+
+ Removing iodbcsrc directory as newer/better versions
+ can be found on the web at www.iodbc.org and
+ www.unixodbc.org.  On Linux, I currently build
+ with unixODBC verson 2.2.0.
+ 
+ Added patch for handling setting the ODBC environment
+ during the connect, thanks to Steffen Goeldner
+ 
+ The attached patch makes it possible to choose the
+ ODBC version. E.g.:
+ 
+   my $dbh = DBI->connect( ..., { odbc_version => 3 } )
+ 
+ directs the driver to exhibit ODBC 3.x behavior.
+
+ Fix to SQLColAttributes thanks to Nicolas DeRico
+
+ Changes to the connect sequence.  If SQLDriverConnect
+ is supported and fails, SQLConnect will be called
+ I<unless> (this is the new part) the length of the
+ DSN is > SQL_MAX_DSN_LENGTH or the DSN begins with
+ DSN=
+
+ New test in mytest which demonstrates long binary
+ types.  If you are having trouble inserting images
+ into your database, please check here.
+=item B<DBD::ODBC 0.38>
+
+ Fixed do function (again) thanks to work by Martin Evans.
+ 
+=item B<DBD::ODBC 0.37>
+
+ Patches for get_info where return type is string.  Patches
+ thanks to Steffen Goldner.  Thanks Steffen!
+
+ Patched get_info to NOT attempt to get data for SQL_DRIVER_HSTMT
+ and SQL_DRIVER_HDESC as they expect data in and have limited value
+ (IMHO).
+
+=item B<DBD::ODBC 0.37>
+
+ Further fixed build for ODBC 2.x drivers.  The new SQLExecDirect
+ code had SQLAllocHandle which is a 3.x function, not a 2.x function.
+ Sigh.  I should have caught that the first time.  Signed, the Mad-and-
+ not-thorough-enough-patcher.
+
+ Additionally, a random core dump occurred in the tests, based upon the
+ new SQLExecDirect code.  This has been fixed.
+ 
+ 
+=item B<DBD::ODBC 0.36>
+
+ Fixed build for ODBC 2.x drivers.  The new SQLExecDirect code
+ had SQLFreeHandle which is a 3.x function, not a 2.x function.
+ 
+=item B<DBD::ODBC 0.35>
+
+ Fixed (finally) multiple result sets with differing
+ numbers of columns.  The final fix was to call
+ SQLFreeStmt(SQL_UNBIND) before repreparing
+ the statement for the next query.
+
+ Added more to the multi-statement tests to ensure
+ the data retrieved was what was expected.
+
+ Now, DBD::ODBC overrides DBI's do to call SQLExecDirect
+ for simple statements (those without parameters).
+ Please advise if you run into problems.  Hopefully,
+ this will provide some small speed improvement for
+ simple "do" statements.  You can also call
+ $dbh->func($stmt, ExecDirect).  I'm not sure this has
+ great value unless you need to ensure SQLExecDirect
+ is being called.  Patches thanks to Merijn Broeren.
+ Thanks Merijn!
+   
+=item B<DBD::ODBC 0.34>
+ 
+ Further revamped tests to attempt to determine if SQLDescribeParam
+ will work to handle the binding types.  The t/08bind.t attempts
+ to determine if SQLDescribeParam is supported.  note that Oracle's
+ ODBC driver under NT doesn't work correctly when binding dates
+ using the ODBC date formatting {d } or {ts }.  So, test #3 will
+ fail in t/08bind.t
+
+ New support for primary_key_info thanks to patches by Martin Evans.
+ New support for catalog, schema, table and table_type in table_info
+ thanks to Martin Evans.  Thanks Martin for your work and your
+ continuing testing, suggestions and general support!
+
+ Support for upcoming dbi get_info.
+ 
 =item B<DBD::ODBC 0.33_3>
 
  Revamped tests to include tests for multiple result sets.
@@ -303,14 +628,18 @@ See L<DBI> for more information.
    "FIX: "Incorrect Syntax near the Keyword 'by' "
    Error Message with Column Names of "C", "CA" or "CAS" (Q273813)
 
- DBD::ODBC now does not name any of the columns A, B, C, or D they are now
- COL_A, COL_B, COL_C, COL_D.
+ DBD::ODBC now does not name any of the columns A, B, C, or D
+ they are now COL_A, COL_B, COL_C, COL_D.
 
- *** NOTE: *** I AM STRONGLY CONSIDERING MAKING THE NEW BINDING the default
-   for future versions.  I do not believe it will break much existing code
-   as anyone binding to non VARCHAR (without the ODBC driver doing a
-   good conversion from the VARCHAR) will have a problem.
-   Please comment soon...
+ *** NOTE: *** I AM STRONGLY CONSIDERING MAKING THE NEW
+ BINDING the default for future versions.  I do not believe
+ it will break much existing code (if any) as anyone binding
+ to non VARCHAR (without the ODBC driver doing a good conversion
+ from the VARCHAR) will have a problem.  It may be subtle, however,
+ since much code will work, but say, binding dates may not with
+ some drivers.
+   
+ Please comment soon...
    
 =item B<DBD::ODBC 0.33_1>
 
@@ -399,7 +728,7 @@ SQL_ROWSET_SIZE attribute patch from Andrew Brown
 > selects are possible.
 >
 > The code change allows for:
-> $dbh->{SQL_ROWSET_SIZE} = 2;    # Any value > 1
+> $dbh->{odbc_SQL_ROWSET_SIZE} = 2;    # Any value > 1
 >
 > For this very purpose.
 >
